import json
import collections
import os
import copy
from functools import partial
from random import choice
import textwrap

import matplotlib as mpl
import matplotlib.pyplot as plt
mpl.use('module://kivy.garden.matplotlib.backend_kivy')
from kivy.garden.matplotlib.backend_kivyagg import FigureCanvasKivyAgg
from kivy.app import App
from kivy.clock import Clock
from kivy.utils import get_color_from_hex
from kivy.core.window import Window
from kivy.animation import Animation
from kivy.uix.actionbar import ActionButton
from kivy.uix.behaviors import FocusBehavior
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.gridlayout import GridLayout
from kivy.uix.textinput import TextInput
from kivy.uix.modalview import ModalView
from kivy.uix.button import Button
from kivy.properties import BooleanProperty, ObjectProperty, StringProperty
from kivy.uix.recycleview import RecycleView
from kivy.uix.recycleview.layout import LayoutSelectionBehavior
from kivy.uix.recycleboxlayout import RecycleBoxLayout
from kivy.uix.recycleview.views import RecycleDataViewBehavior
from kivy.uix.popup import Popup
from kivy.uix.spinner import SpinnerOption, Spinner
from kivy.uix.label import Label
from kivy.uix.screenmanager import Screen
from kivy.uix.togglebutton import ToggleButton

from es_gui.proving_grounds.help_carousel import HelpCarouselModalView

cwd = os.getcwd()

# Animation durations in seconds #
BASE_TRANSITION_DUR = 0.600
BUTTON_FLASH_DUR = 0.100
ANIM_STAGGER = 0.200
FADEIN_DUR = 1.000
SLIDER_DUR = 1.200
LOADING_DUR = 1.000

# Widget sizes #
TWO_ABC_WIDTH = 800
THREE_ABC_WIDTH = 1200

# Font sizes #
default_font = 20
stnd_font = 22
mid_font = 25
large_font = 30
huge_font = 40

# Software name and information #
APP_NAME = 'QuESt'
APP_TAGLINE = 'Optimizing Energy Storage'

# Palette #
PALETTE = [(0, 83, 118), (132, 189, 0),
     (0, 173, 208), (255, 163, 0), (255, 88, 93), (174, 37, 115)]

def rgba_to_fraction(rgba):
    """Converts rgb values in int format to fractional values suitable for Kivy."""
    if len(rgba) > 3:
        return float(rgba[0])/255, float(rgba[1])/255, float(rgba[2])/255, rgba[3]
    else:
        return float(rgba[0])/255, float(rgba[1])/255, float(rgba[2])/255, 1

def fade_in_animation(content, *args):
    """Fade in animation (on opacity); to be used with Clock scheduler."""
    anim = Animation(transition='out_expo', duration=FADEIN_DUR, opacity=1)
    anim.start(content)

def slow_blinking_animation(content, *args):
    """Slow blinking animation (on opacity); to be used with Clock scheduler."""
    anim = Animation(transition='linear', duration=LOADING_DUR, opacity=0) + Animation(transition='linear', duration=LOADING_DUR, opacity=1)
    anim.repeat = True
    anim.start(content)


class NavigationButton(ActionButton):
    pass


class LeftAlignedText(Label):
    """Label subclass for left-aligned text labels."""
    pass


class BodyTextBase(LeftAlignedText):
    """Base class for body text labels."""
    pass


class TitleTextBase(LeftAlignedText):
    """Base class for title text labels."""
    pass


class TileButton(Button):
    """Button subclass for the tile buttons used throughout the application."""
    pass


class MenuTileButton(TileButton):
    """Large tile button used for main menus."""
    pass


class SelectableRecycleBoxLayout(FocusBehavior, LayoutSelectionBehavior,
                                 RecycleBoxLayout):
    """The BoxLayout for displaying RecycleView. Adds selection and focus behavior to the RecycleView."""
    pass


class RecycleViewRow(RecycleDataViewBehavior, BoxLayout):
    """The representation for data entries in the RecycleView. A selectable Label with the name of the saved run."""
    index = None
    selected = BooleanProperty(False)
    selectable = BooleanProperty(True)

    # run_selector = None

    def refresh_view_attrs(self, rv, index, data):
        """
        Catch and handle the view changes.
        """
        self.index = index
        return super(RecycleViewRow, self).refresh_view_attrs(
            rv, index, data)

    def on_enter(self):
        """Update the description boxes when hovered over."""
        pass
        # run_obj = self.run_selector.rv.data[self.index]
        # run_desc_box = self.run_selector.rv_desc
        #
        # # Update the preview text based on the hovered upon object.
        # run_desc_box.text = '\n'.join(
        #     ['[b]{0}[/b]'.format(run_obj['name']),
        #      '[b]Date: [/b] ' + run_obj['time'],
        #      '[b]ISO:[/b] ' + run_obj['iso'],
        #      '[b]Market Type:[/b] ' + run_obj['market type'],
        #      '[b]Year:[/b] ' + run_obj['year'],
        #      '[b]Month:[/b] ' + run_obj['month'],
        #      '[b]Node:[/b] ' + run_obj['node'],
        #      '[b]Parameters set:[/b] ' + repr(run_obj.get('params', 'Used defaults.')),
        #      ]
        # )

    def on_leave(self):
        """
        Update the description boxes when not hovering over an entry.
        """
        pass
        # run_desc_box = self.run_selector.rv_desc
        #
        # # update the text
        # run_desc_box.text = 'Hover over a saved run to view its details.'

    def on_touch_down(self, touch):
        """
        Add selection on touch down.
        """
        if super(RecycleViewRow, self).on_touch_down(touch):
            return True
        if self.collide_point(*touch.pos) and self.selectable:
            return self.parent.select_with_touch(self.index, touch)

    def apply_selection(self, rv, index, is_selected):
        """
        Respond to the selection of items in the view.
        """
        self.selected = is_selected


class MyRecycleView(RecycleView):
    def __init__(self, **kwargs):
        super(MyRecycleView, self).__init__(**kwargs)

        self.unfiltered_data = self.data

    def filter_rv_data(self, filter_text):
        self.deselect_all_nodes()
                
        if filter_text:
            self.data = [rv_entry for rv_entry in self.unfiltered_data if filter_text.lower() in rv_entry['name'].lower()]
        else:
            self.data = self.unfiltered_data
    
    def deselect_all_nodes(self):
        while len(self.layout_manager.selected_nodes) > 0:
            for selected_node in self.layout_manager.selected_nodes:
                self.layout_manager.deselect_node(selected_node)


class MyPopup(Popup):
    pass


class WarningPopup(MyPopup):
    def __init__(self, **kwargs):
        super(WarningPopup, self).__init__(**kwargs)

        self._keyboard = Window.request_keyboard(self._keyboard_closed, self, 'text')

        if self._keyboard.widget:
            pass

        self._keyboard.bind(on_key_down=self._on_keyboard_down)

    def _keyboard_closed(self):
        self._keyboard.unbind(on_key_down=self._on_keyboard_down)
        self._keyboard = None

    def _on_keyboard_down(self, keyboard, keycode, text, modifiers):
        if keycode[1] in ('enter', 'numpadenter'):
            self.dismiss()

        return True


class ValuationRunCompletePopup(MyPopup):
    pass


class ConnectionErrorPopup(MyPopup):
    def open_connection_settings(self):
        """Opens the settings screen."""
        settings_screen = App.get_running_app().settings.parent.parent
        settings_screen.open()


class PlotSpinner(Spinner, TileButton):
    pass


class PlotSpinnerOptionCls(SpinnerOption):
    pass


class InputError(Exception):
    pass


class LoadingModalView(ModalView):
    pass
    # def on_open(self):
    #     loading_animation = Animation(transition='linear', duration=LOADING_DUR, opacity=0) + Animation(transition='linear', duration=LOADING_DUR, opacity=1)

    #     for x in range(5):
    #         loading_animation += Animation(transition='linear', duration=LOADING_DUR, opacity=0) + Animation(transition='linear', duration=LOADING_DUR, opacity=1)
    #     # loading_animation.repeat = True

    #     Clock.schedule_once(lambda dt: loading_animation.start(self.logo), 0)


class ReportScreen(Screen):
    pass


class WizardCompletePopup(MyPopup):
    pass


class WizardReportInterface(Screen):
    def on_enter(self):
        def _start(*args):
            self.chart_type_toggle.children[-1].state = 'down'

            # random_report = choice(self.chart_type_toggle.children)
            # random_report.state = 'down'

        if not any([button.state == 'down' for button in self.chart_type_toggle.children]):
            Clock.schedule_once(lambda dt: _start(), 0.25)


class ReportChartToggle(ToggleButton, TileButton):
    pass


class ParameterRow(GridLayout):
    """Grid layout containing parameter name, description, text input, and units."""
    def __init__(self, desc, **kwargs):
        super(ParameterRow, self).__init__(**kwargs)

        self._desc = desc

        self.name.text = self.desc.get('name', '')
        self.notes.text = self.desc.get('notes', '')
        self.text_input.hint_text = str(self.desc.get('default', ''))
        self.units.text = self.desc.get('units', '')

    @property
    def desc(self):
        return self._desc

    @desc.setter
    def desc(self, value):
        self._desc = value


class ParameterRow2cols(GridLayout):
    """Grid layout containing parameter name and text input."""
    def __init__(self, desc, **kwargs):
        super(ParameterRow2cols, self).__init__(**kwargs)

        self._desc = desc

        self.name.text = self.desc.get('name', '')
        self.text_input.text = str(self.desc.get('value', ''))

    @property
    def desc(self):
        return self._desc

    @desc.setter
    def desc(self, value):
        self._desc = value


class ParamTextInput(TextInput):
    """A TextInput field for entering parameter values. Limited to float values."""
    def insert_text(self, substring, from_undo=False):
        # limit to 8 chars
        substring = substring[:8 - len(self.text)]
        return super(ParamTextInput, self).insert_text(substring, from_undo=from_undo)
    
class ParameterRowText(GridLayout):
    """Grid layout containing parameter name, description, text input, and units."""
    def __init__(self, desc, **kwargs):
        super(ParameterRowText, self).__init__(**kwargs)

        self._desc = desc

        self.name.text = self.desc.get('name', '')
        self.notes.text = self.desc.get('notes', '')
        self.text_input.hint_text = str(self.desc.get('default', ''))
        self.units.text = self.desc.get('units', '')

    @property
    def desc(self):
        return self._desc

    @desc.setter
    def desc(self, value):
        self._desc = value


class ParamTextInputText(TextInput):
    """A TextInput field for entering parameter values. Limited to float values."""
    def insert_text(self, substring, from_undo=False):
        # limit to 25 chars
        substring = substring[:25 - len(self.text)]
        return super(ParamTextInputText, self).insert_text(substring, from_undo=from_undo)


class ParameterGridWidget(GridLayout):
    """Grid layout containing rows of parameter adjustment widgets."""   
    def _validate_inputs(self):
        params = []
        param_set = {}

        for row in self.children:
            attr_name = row.desc['attr name']

            if not row.text_input.text:
                attr_val = row.text_input.hint_text
            else:
                attr_val = row.text_input.text
            
            param_set[attr_name] = float(attr_val)
        
        params.append(param_set)

        return params
    
    def get_inputs(self):
        try:
            params = self._validate_inputs()
        except InputError as e:
            popup = WarningPopup()
            popup.popup_text.text = str(e)
            popup.open()
        else:
            return params
    
    def get_input_strings(self):
        params = []

        for row in self.children:
            param_name = row.desc['name']
            param_units = row.desc['units']
            
            if not row.text_input.text:
                param_val = row.text_input.hint_text
            else:
                param_val = row.text_input.text
            
            param_string = '{name}: {value} {units}'.format(name=param_name, value=param_val, units=param_units)
            params.append(param_string)
        
        return params


class DataGovAPIhelp(ModalView):
    """ModalView to display instructions on how to get a Data.gov (NREL Developer Network) API key."""


class ResultsViewer(Screen):
    """The screen for displaying plots inside the application or exporting results."""
    current_fig = ObjectProperty()
    current_ax = ObjectProperty()

    def __init__(self, **kwargs):
        super(ResultsViewer, self).__init__(**kwargs)

        self.dfs = {}

        self.time_selector.start_time.bind(on_text_validate=self.draw_figure)
        self.time_selector.end_time.bind(on_text_validate=self.draw_figure)
    
    def on_enter(self):
        ab = self.manager.nav_bar
        ab.set_title('Results Viewer')

        help_button = NavigationButton(
            text='help (results viewer)',
            on_release=self.open_help_carousel,
        )

        ab.action_view.add_widget(help_button)
    
    def open_help_carousel(self, *args):
        """
        """
        help_carousel_view = HelpCarouselModalView()
        help_carousel_view.title.text = "Results Viewer"

        slide_01_text = "The Results Viewer is a built-in tool to help you look at QuESt optimization results. The recycle view at the bottom contains each optimization run (model) performed during your current session. Typically each item will correspond to a single month.\n\nYou can select which models you want to view simultaneously. We recommend selecting no more than six at a time."

        slide_02_text = "The 'Select data' spinner is for selecting which quantity to plot. The variety of choices here will differ among QuESt applications. While most selections correspond to line plots of time series, some other plots such as box-and-whisker plots may be available."

        slide_03_text = "Click on the 'Plot/Redraw' to render the plot according to your current selections.\n\nNote that the figure is not interactive."

        slide_04_text = "For time series plots, you can adjust the range of time shown using the 'Hours shown' field. This can be used to look at specific points in time in more detail.\n\nYou can hit the 'Enter' key after changing these values to quickly render the plot."

        slide_05_text = "You can export the currently rendered plot to a PNG image file using the 'Export PNG' button.\n\nYou can also export the detailed table of results for each selected model to a CSV file using the 'Export CSV' button. An individual file will be made for each selected model. These files contain details such as decision variable values at each timestep and other relevant quantities."

        slides = [
            (os.path.join("es_gui", "resources", "help_views", "common", "results_viewer", "01.png"), slide_01_text),
            (os.path.join("es_gui", "resources", "help_views", "common", "results_viewer", "02.png"), slide_02_text),
            (os.path.join("es_gui", "resources", "help_views", "common", "results_viewer", "03.png"), slide_03_text),
            (os.path.join("es_gui", "resources", "help_views", "common", "results_viewer", "04.png"), slide_04_text),
            (os.path.join("es_gui", "resources", "help_views", "common", "results_viewer", "05.png"), slide_05_text),
        ]

        help_carousel_view.add_slides(slides)
        help_carousel_view.open()

    def on_pre_enter(self):
        pass

    def on_leave(self):
        """Resets all selections and the graph."""
        self._reset_screen()

    # def _on_keyboard_down(self, instance, keyboard, keycode, text, modifiers):
    #     if keycode == 40:  # 40 - Enter key pressed
    #         self.draw_figure()

    def _select_all_runs(self):
        for selectable_node in self.rv.layout_manager.get_selectable_nodes():
            self.rv.layout_manager.select_node(selectable_node)

    def _deselect_all_runs(self):
        while len(self.rv.layout_manager.selected_nodes) > 0:
            for selected_node in self.rv.layout_manager.selected_nodes:
                self.rv.layout_manager.deselect_node(selected_node)

    def _reset_screen(self):
        #rv = self.run_selector.rv
        rv = self.rv

        self._deselect_all_runs()

        # Clears graph.
        while len(self.plotbox.children) > 0:
            for widget in self.plotbox.children:
                if isinstance(widget, FigureCanvasKivyAgg):
                    self.plotbox.remove_widget(widget)

        # Reset spinners.
        self.vars_button.text = 'Select data'

        # Reset text input fields.
        self.time_selector.start_time.text = '0'
        self.time_selector.end_time.text = '744'

    def _reinit_graph(self, has_legend=True):
        # Clears graph.
        while len(self.plotbox.children) > 0:
            for widget in self.plotbox.children:
                if isinstance(widget, FigureCanvasKivyAgg):
                    self.plotbox.remove_widget(widget)
        
        fig, ax = plt.subplots()
        self.current_fig = fig
        self.current_ax = ax

        ax.xaxis.label.set_size(16)
        ax.yaxis.label.set_size(16)
        plt.rcParams.update({'font.size': 18, 'xtick.labelsize': 12, 'ytick.labelsize': 12,
                             'lines.linewidth': 2})
        plt.rc('legend', **{'fontsize': 10})

        if has_legend:
            box = ax.get_position()
            ax.set_position([box.x0, box.y0, box.width * 0.75, box.height])

        plotbox = self.plotbox
        canvas = FigureCanvasKivyAgg(self.current_fig)
        plotbox.add_widget(canvas, index=0)

    def _update_toolbar(self, *args):
        """Updates the data viewing toolbar based on selections."""
        # if not self.dfs:
        #     # if no selections made, disable all toolbar buttons and return
        #     self.vars_button.disabled = True
        #     self.time_button.disabled = True
        #     self.draw_button.disabled = True
        #     self.csv_export_button.disabled = True
        #     self.png_export_button.disabled = True
        #
        #     return

        self.vars_button.disabled = False
        #self.time_button.disabled = False
        self.draw_button.disabled = False
        self.csv_export_button.disabled = False
        self.png_export_button.disabled = False

    def _update_selection(self):
        """Updates the dict. of DataFrames whenever new selections of data are made."""

        # Identify the selected run(s) from RunSelector.
        #rv = self.run_selector.rv
        rv = self.rv
        runs_selected = [rv.data[selected_ix] for selected_ix in rv.layout_manager.selected_nodes]

        results = {}

        for run in runs_selected:
            label = run['name']
            df = run['optimizer'].results

            results[label] = df

        self.dfs = results

    def _validate_inputs(self):
        if not self.dfs:
            # No selections made.
            popup = WarningPopup()
            popup.popup_text.text = "We need data to plot! Let's pick some solved models to view first."
            popup.open()

            return False

        plot_type = self.vars_button.text

        if plot_type == 'Select data':
            # No data to plot has been selected.
            popup = WarningPopup()
            popup.popup_text.text = "We need something to plot! Let's select some data to view first."
            popup.open()

            return False

        if not self.time_selector.validate():
            return False

        return True

    def draw_figure(self, *args):
        pass

    def export_png(self, outdir_root):
        """Exports currently displayed figure to .png file in specified location."""
        os.makedirs(outdir_root, exist_ok=True)

        self._update_selection()

        if not self.plotbox.children:
            popup = WarningPopup()
            popup.popup_text.text = "There is currently no plot drawn to export."
            popup.open()
        elif self.dfs:
            outname = os.path.join(outdir_root, self.vars_button.text+'.png')
            self.plotbox.children[0].export_to_png(outname)

            popup = WarningPopup(size_hint=(0.4, 0.4))
            popup.title = 'Success!'
            popup.popup_text.text = 'Figure successfully exported to:\n\n' + os.path.abspath(outname)
            popup.open()
        else:
            popup = WarningPopup()
            popup.popup_text.text = "We need a plot to export! Let's pick some solved models to view first."
            popup.open()

    def export_csv(self, outdir_root):
        """Exports selected DataFrames to .csv files in specified location."""
        os.makedirs(outdir_root, exist_ok=True)
        
        self._update_selection()

        if self.dfs:
            for run in self.dfs:
                # Split and regenerate run label to make it filename-friendly.
                run_label_split = run.split(' | ')
                run_name = ' '.join(run_label_split[:4])

                # Strip non-alphanumeric chars.
                delchars = ''.join(c for c in map(chr, range(256)) if not c.isalnum())
                run_name = run_name.translate({ord(i): None for i in delchars})

                outname = os.path.join(outdir_root, run_name + '.csv')
                df = self.dfs[run]
                os.makedirs(outdir_root, exist_ok=True)
                df.to_csv(outname, index=False)

            popup = WarningPopup(size_hint=(0.4, 0.4))
            popup.title = 'Success!'
            popup.popup_text.text = 'File(s) successfully exported to:\n\n' + os.path.abspath(outdir_root)
            popup.open()
        else:
            popup = WarningPopup()
            popup.popup_text.text = "We need some results to export! Let's pick some solved models to view first."
            popup.open()


class TimeSelector(MyPopup):
    """
    A Popup with TextInput fields for determining the range of time to display in the figure.
    """

    def __init__(self, **kwargs):
        super(TimeSelector, self).__init__(**kwargs)

    def _validate(self):
        """
        Validates the field entries before closing the TimeSelector.
        """
        try:
            start_time = int(self.start_time.text)
            end_time = int(self.end_time.text)
            assert(start_time < end_time)
        except ValueError:
            # empty text input
            popup = WarningPopup()
            popup.popup_text.text = 'All input fields must be populated to continue.'
            popup.open()
        except AssertionError:
            # start_time > end_time
            popup = WarningPopup()
            popup.popup_text.text = 'The start time cannot be greater than the end time.'
            popup.open()
        else:
            self.dismiss()

        # Note: We do not necessarily need to check for index errors because Series slicing does not throw exceptions


class TimeSelectorRow(GridLayout):
    def get_inputs(self):
        return int(self.start_time.text), int(self.end_time.text)

    def validate(self):
        """Validates the field entries before closing the TimeSelector."""
        try:
            start_time = int(self.start_time.text)
            end_time = int(self.end_time.text)
            assert(start_time < end_time)
        except ValueError:
            # empty text input
            popup = WarningPopup()
            popup.popup_text.text = 'All input fields must be populated to continue.'
            popup.open()

            return False
        except AssertionError:
            # start_time > end_time
            popup = WarningPopup()
            popup.popup_text.text = 'The start time cannot be greater than the end time.'
            popup.open()

            return False
        else:
            return True


class TimeTextInput(TextInput):
    """
    A TextInput field for entering time indices. Limited to three int characters only.
    """
    def insert_text(self, substring, from_undo=False):
        # limit to 3 chars
        substring = substring[:3 - len(self.text)]
        return super(TimeTextInput, self).insert_text(substring, from_undo=from_undo)


class ValuationParameterRow(GridLayout):
    """Grid layout containing parameter descriptor label and text input field. For QuESt Valuation interfaces."""
    def __init__(self, desc, **kwargs):
        super(ValuationParameterRow, self).__init__(**kwargs)

        self._desc = desc
        self.name.text = self.desc['name']
        self.text_input.hint_text = str(self.desc['default'])

    @property
    def desc(self):
        return self._desc

    @desc.setter
    def desc(self, value):
        self._desc = value


class ValuationParameterWidget(GridLayout):
    """Grid layout containing rows of ValuationParameterRow widgets. For inputting parameter values into text input fields in QuESt Valuation interfaces."""
    def build(self, iso):
        # Build the widget by creating a row for each parameter.
        data_manager = App.get_running_app().data_manager
        MODEL_PARAMS = data_manager.get_valuation_model_params(iso)

        for param in MODEL_PARAMS:
            row = ValuationParameterRow(desc=param)
            self.add_widget(row)
            setattr(self, param['attr name'], row)
    
    def validate_inputs(self):
        # Check the parameter inputs.
        param_dict = {}

        # Check for any input into the parameter rows.
        for param_row in self.children:
            param_name = param_row.name.text
            attr_name = param_row.desc['attr name']

            if param_row.text_input.text:
                param_value = float(param_row.text_input.text)
            else:
                # Use the hint text (default value)
                param_value = float(param_row.text_input.hint_text)

            param_dict[attr_name] = param_value

            # Values cannot be negative.
            if param_value < 0:
                raise InputError('"{0}" cannot be negative. (got {1})'.format(param_name, param_value))
            
            # Percentages cannot exceed 100.
            if attr_name in {'Self_discharge_efficiency', 'Round_trip_efficiency', 'State_of_charge_init', 'State_of_charge_min', 'State_of_charge_max', 'Reserve_reg_min', 'Reserve_reg_max',} and param_value > 100:
                raise InputError('"{0}" cannot exceed 100%. (got {1})'.format(param_name, param_value))

        # Minimum state of charge must be strictly less than the maximum state of charge.
        if not param_dict['State_of_charge_max'] > param_dict['State_of_charge_min']:
            raise InputError('The maximum state of charge must be greater than the minimum state of charge.')

        # Initial state of charge must be between the minimum and maximum state of charge values.
        if not (param_dict['State_of_charge_max'] >= param_dict['State_of_charge_init'] and 
        param_dict['State_of_charge_init'] >= param_dict['State_of_charge_min']):
            raise InputError('The initial state of charge must be between the minimum and maximum state of charge values.')
    
    def get_inputs(self, use_hint_text=False):
        self.validate_inputs()

        base_param_dict = {}

        # Check for any input into the parameter rows.
        for param_row in self.children:
            param_name = param_row.name.text
            attr_name = param_row.desc['attr name']

            if not param_row.text_input.text and not use_hint_text:
                continue
            elif param_row.text_input.text:
                param_value = float(param_row.text_input.text)
            else:
                param_value = float(param_row.text_input.hint_text)
            
            base_param_dict[attr_name] = param_value

        return base_param_dict


class ValuationParamTextInput(TextInput):
    """A TextInput field for entering parameter value sweep range descriptors. Limited to float values."""
    def insert_text(self, substring, from_undo=False):
        # limit to 8 chars
        substring = substring[:8 - len(self.text)]
        return super(ValuationParamTextInput, self).insert_text(substring, from_undo=from_undo)
    
class PerformanceParameterRow(GridLayout):
    """Grid layout containing parameter descriptor label and text input field. For QuESt Valuation interfaces."""
    def __init__(self, lbl, desc, **kwargs):
        super(PerformanceParameterRow, self).__init__(**kwargs)

        self._lbl = lbl
        self._desc = desc
        self.name.text = self.desc[0]
        self.text_input.hint_text = str(self.desc[1])

    @property 
    def lbl(self):
        return self._lbl
    
    @lbl.setter
    def lbl(self, value):
        self._lbl = value
        
    @property
    def desc(self):
        return self._desc

    @desc.setter
    def desc(self, value):
        self._desc = value
    
class PerformanceParameterWidget(GridLayout):
    """Grid layout containing rows of ValuationParameterRow widgets. For inputting parameter values into text input fields in QuESt Valuation interfaces."""
    def build(self):
        # Build the widget by creating a row for each parameter.
        data_manager = App.get_running_app().data_manager
        MODEL_PARAMS = {
                'eCap': ['Energy Capacity (MWh)', 1],
                'pRat': ['Power Rating (MW)', 1],
                'n_s' : ['Self-Discharge Efficiency', 1],
                'n_p' : ['Power Electronics Efficiency', 0.93],
                'q_rate': ['Battery Cell Ah Rating', 2.5],
                'v_rate':['Battery Cell Voltage Rating', 3.6],
                'r': ['Battery Cell Internal Resistance',0.02],
                'k': ['Battery Cell k Parameter',0.005],
                'tau': ['Time Step (hr)', 0.25],
                'h_setpoint': ['Heating Setpoint (C)', 15],
                'c_setpoint': ['Cooling Setpoint (C)', 40],
                'insulation': ['Insulation (m\u00b2 K/W)', 0]
                }

        for param in MODEL_PARAMS:
            if not param == 'tau':
                row = PerformanceParameterRow(lbl = param, desc=MODEL_PARAMS[param])
                self.add_widget(row)
                setattr(self, param[0], row)
            
        return True
    
    def validate_inputs(self):
        # Check the parameter inputs.
        param_dict = {}

        # Check for any input into the parameter rows.
        for param_row in self.children:
            param_name = param_row.name.text
            attr_name = param_row.desc[0]

            if param_row.text_input.text:
                param_value = float(param_row.text_input.text)
            else:
                # Use the hint text (default value)
                param_value = float(param_row.text_input.hint_text)

            param_dict[attr_name] = param_value

            # Values cannot be negative.
            if param_value < 0:
                raise InputError('"{0}" cannot be negative. (got {1})'.format(param_name, param_value))
            
            # Percentages cannot exceed 100.
            if attr_name in {'Self_discharge_efficiency', 'Round_trip_efficiency', 'State_of_charge_init', 'State_of_charge_min', 'State_of_charge_max', 'Reserve_reg_min', 'Reserve_reg_max',} and param_value > 100:
                raise InputError('"{0}" cannot exceed 100%. (got {1})'.format(param_name, param_value))
    
    def get_inputs(self, use_hint_text=False):
        self.validate_inputs()

        base_param_dict = {}

        # Check for any input into the parameter rows.
        for param_row in self.children:
            param_name = param_row.name.text
            attr_name = param_row.lbl
            
            if not param_row.text_input.text and not use_hint_text:
                continue
            elif param_row.text_input.text:
                param_value = float(param_row.text_input.text)
            else:
                param_value = float(param_row.text_input.hint_text)
            
            base_param_dict[attr_name] = param_value
            

        return base_param_dict


class PerformanceParamTextInput(TextInput):
    """A TextInput field for entering parameter value sweep range descriptors. Limited to float values."""
    def insert_text(self, substring, from_undo=False):
        # limit to 8 chars
        substring = substring[:8 - len(self.text)]
        return super(ValuationParamTextInput, self).insert_text(substring, from_undo=from_undo)



from __future__ import absolute_import, print_function

import os
import json
import re

from kivy.uix.screenmanager import Screen, ScreenManager, SlideTransition
from kivy.uix.gridlayout import GridLayout
from kivy.properties import BooleanProperty
from kivy.app import App

from es_gui.resources.widgets.common import ValuationRunCompletePopup, RecycleViewRow, BASE_TRANSITION_DUR, ParameterRow2cols
from es_gui.apps.tech_selection.results_viewer import TechSelectionFeasible
from es_gui.apps.tech_selection.analysis import perform_tech_selection
from es_gui.apps.tech_selection.fAux import pdSeriesIdxWhereTrue


class TechSelectionWizard(Screen):
    """The main screen for the technology selection wizard. This hosts the nested screen manager for the actual wizard."""

    def on_enter(self):
        """Create navigation bar and screen title."""
        ab = self.manager.nav_bar
        ab.set_title('Energy Storage Technology Selection Application')

    def on_leave(self):
        """Reset wizard to initial state by removing all screens except the first."""
        self.sm.current = 'tech_selection_home'
        if len(self.sm.screens) > 1:
            self.sm.clear_widgets(screens=self.sm.screens[1:])


class TechSelectionWizardScreenManager(ScreenManager):
    """The screen manager for the technology selection wizard screens."""

    def __init__(self, **kwargs):
        super(TechSelectionWizardScreenManager, self).__init__(**kwargs)
        self.transition = SlideTransition()
        self.add_widget(TechSelectionWizardStart(name='tech_selection_home'))


class TechSelectionWizardStart(Screen):
    """The starting/welcome screen for the technology selection wizard."""

    def _next_screen(self):
        """Set actions for when the 'Next' button is pressed."""

        # Create, if necessary, the next screen
        if not self.manager.has_screen('tech_selection_inputs'):
            screen = TechSelectionUserInputs(name='tech_selection_inputs')
            self.manager.add_widget(screen)

        # Move to the next screen
        self.manager.transition.duration = BASE_TRANSITION_DUR
        self.manager.transition.direction = 'left'
        self.manager.current = 'tech_selection_inputs'


class TechSelectionUserInputs(Screen):
    """The user inputs screen for the technology selection wizard."""

    has_selection = BooleanProperty(False) # Boolean variable indicating whether user selections have been made

    def __init__(self, **kwargs):
        super(TechSelectionUserInputs, self).__init__(**kwargs)
        self.selected_params = {} # Dictionary to collect/save user selections

        # 'Children screens' for each RecycleView
        GridLocationRVEntry.host_screen = self
        ApplicationRVEntry.host_screen = self
        SystemSizeRVEntry.host_screen = self
        DischargeDurationRVEntry.host_screen = self

        self.data_manager = App.get_running_app().data_manager

        self.target_cost_kW = 1500
        self.target_cost_kWh = 1000

        # Read input databases
        self.apps_db = self.data_manager.get_applications_db()
        self.all_user_options_names = self.data_manager.get_tech_selection_all_options()
        self.all_apps = self.apps_db.index.to_list()

        self.default_inputs = {
            'Transmission/central':
                [pdSeriesIdxWhereTrue(self.apps_db['Default at transmission'] == 'Yes')[0], 'Wholesale (100 MW)'],
            'Distribution':
                [pdSeriesIdxWhereTrue(self.apps_db['Default at distribution'] == 'Yes')[0], 'Distribution & microgrid (1 MW)'],
            'BTM: commercial/industrial':
                [pdSeriesIdxWhereTrue(self.apps_db['Default at industrial'] == 'Yes')[0], 'Commercial & industrial (100 kW)'],
            'BTM: residential':
                [pdSeriesIdxWhereTrue(self.apps_db['Default at residential'] == 'Yes')[0], 'Residential (10 kW)']}

        self.compatibility_apps = {
            'Transmission/central': pdSeriesIdxWhereTrue(self.apps_db['Compatible with transmission'] == 'Yes'),
            'Distribution': pdSeriesIdxWhereTrue(self.apps_db['Compatible with distribution'] == 'Yes'),
            'BTM: commercial/industrial': pdSeriesIdxWhereTrue(self.apps_db['Compatible with industrial'] == 'Yes'),
            'BTM: residential': pdSeriesIdxWhereTrue(self.apps_db['Compatible with residential'] == 'Yes')}

        # Populate lists for each category of user input
        self.ids.grid_location_rv.data = [{'name': value} for value in self.all_user_options_names['grid_location']]
        self.ids.app_names_rv.data = [{'name': value} for value in self.all_apps]
        self.ids.system_size_rv.data = [{'name': value} for value in self.all_user_options_names['system_size']]
        self.ids.discharge_duration_rv.data = [{'name': value} for value in self.all_user_options_names['discharge_duration']]

    def on_pre_enter(self):
        """Update list of available user selections according to previous selection."""

        # If a grid location has been selected previously, update the list of available applications accordingly
        try:
            self.set_compatible_apps()
        except:
            pass

    def on_has_selection(self, instance, value):
        """Enable the 'Next' button only if all user inputs have been selected."""
        self.next_button.disabled = not value

    def is_user_selection_complete(self):
        """Check if all user selections have been made and return True; otherwise, return False."""
        return all([hasattr(GridLocationRVEntry.host_screen, 'grid_location_selected'),
                    hasattr(ApplicationRVEntry.host_screen, 'application_selected'),
                    hasattr(SystemSizeRVEntry.host_screen, 'system_size_selected'),
                    hasattr(DischargeDurationRVEntry.host_screen, 'discharge_duration_selected')])

    def set_compatible_apps(self):
        """Update the list of available application names based on the selected grid location."""
        self.available_apps_at_location = sorted(self.compatibility_apps[
            GridLocationRVEntry.host_screen.grid_location_selected])
        self.ids.app_names_rv.data = [{'name': value} for value in self.available_apps_at_location]

    def set_default_inputs_app_and_size(self):
        """Set default values for application and system size based on the selected grid location."""
        selected_location = GridLocationRVEntry.host_screen.grid_location_selected
        default_application, default_size = self.default_inputs[selected_location]
        self.ids.app_names_rv_selectable.selected_nodes = [self.available_apps_at_location.index(default_application)]
        self.ids.system_size_rv_selectable.selected_nodes = [self.all_user_options_names['system_size'].index(default_size)]

    def set_default_inputs_duration(self):
        """Set default value for discharge duration based on the selected application."""
        selected_application = ApplicationRVEntry.host_screen.application_selected
        default_duration = ('4 hrs' if self.apps_db.loc[selected_application, 'Power vs. Energy'] == 'Energy'
                            else 'Up to 0.5 hr')
        self.ids.discharge_duration_rv_selectable.selected_nodes = [
            self.all_user_options_names['discharge_duration'].index(default_duration)]

    def _next_screen(self, *args):
        """Set actions for when the 'Next' button is pressed."""

        # Collect/save user selections in a .json file
        self.selected_params = [
            {'name': 'Grid location', 'attr name': 'location',
             'value': GridLocationRVEntry.host_screen.grid_location_selected},
            {'name': 'Application', 'attr name': 'application',
             'value': ApplicationRVEntry.host_screen.application_selected},
            {'name': 'System size', 'attr name': 'system_size',
             'value': re.search(r'\((.*?)\)', SystemSizeRVEntry.host_screen.system_size_selected).group(1)},
            {'name': 'Discharge duration', 'attr name': 'discharge_duration',
             'value': DischargeDurationRVEntry.host_screen.discharge_duration_selected},
            {'name': 'Type of application', 'attr name': 'app_type',
             'value': self.apps_db.loc[ApplicationRVEntry.host_screen.application_selected, 'Power vs. Energy']}]
        destination_file = os.path.join('es_gui', 'apps', 'data_manager', '_static', 'tech_selection_params.json')
        with open(destination_file, 'w') as outfile:
            json.dump(self.selected_params, outfile)

        # Create, if necessary, the next screen
        if not self.manager.has_screen('confirm_inputs'):
            screen = TechSelectionConfirmInputs(name='confirm_inputs')
            self.manager.add_widget(screen)

        # Move to the next screen
        self.manager.transition.duration = BASE_TRANSITION_DUR
        self.manager.transition.direction = 'left'
        self.manager.current = 'confirm_inputs'


class TechSelectionConfirmInputs(Screen):
    """The user inputs confirmation screen for the technology selection wizard."""

    def __init__(self, **kwargs):
        super(TechSelectionConfirmInputs, self).__init__(**kwargs)

    def on_pre_enter(self):
        """Clear any widgets already present in the screen and create a new widget displaying the user selections."""
        self.data_manager = App.get_running_app().data_manager

        if self.param_widget.children:
            self.param_widget.clear_widgets()
        self.param_widget.build()

    def execute_run(self):
        """Run feasibility scores computation according to the user inputs."""

        # Collect user selections values from a .json file
        MODEL_PARAMS = self.data_manager.get_tech_selection_params()
        user_selections = {val['attr name']: val['value'] for val in MODEL_PARAMS}

        dfFeasible, dfRanking = perform_tech_selection(user_selections,
                                                       self.manager.get_screen('tech_selection_inputs').target_cost_kWh,
                                                       self.manager.get_screen('tech_selection_inputs').target_cost_kW)

        dfRanking.to_csv(os.path.join('results', 'tech_selection', 'table_ranking.csv'))

         # Popup window after analysis has been completed
        popup = ValuationRunCompletePopup(size_hint=(0.4, 0.37))
        popup.title = 'Success!'
        popup.popup_text.text = ('The analysis has been completed successfully. Click [i]Go back[/i] to return to the '
                                 'inputs selection screen or [i]Continue[/i] to see the results.')
        popup.view_results_button.text = 'Go back'
        popup.view_results_button.bind(on_release=popup.dismiss)
        popup.dismiss_button.text = 'Continue'
        popup.dismiss_button.bind(on_release=self._next_screen)
        popup.open()

    def _next_screen(self, *args):
        """Set actions for when the 'Next' button is pressed."""

        # Create, if necessary, the next screen
        if not self.manager.has_screen('feasible_techs'):
            screen = TechSelectionFeasible(name='feasible_techs')
            self.manager.add_widget(screen)

        # Move to the next screen
        self.manager.transition.duration = BASE_TRANSITION_DUR
        self.manager.transition.direction = 'left'
        self.manager.current = 'feasible_techs'


class TechSelectionParameterWidget(GridLayout):
    """Grid layout containing rows of user selections widgets."""

    def build(self):
        """Build the widget by creating a row for each parameter."""

        # Collect user selections values from a .json file
        data_manager = App.get_running_app().data_manager
        MODEL_PARAMS = data_manager.get_tech_selection_params()

        # Create and add to the screen one widget for each type of user input
        for param in MODEL_PARAMS:
            if param['attr name'] != 'app_type':
                row = ParameterRow2cols(desc=param)
                self.add_widget(row)


class GridLocationRVEntry(RecycleViewRow):
    """The representation for data entries in the RecycleView for 'Grid Location'."""

    def apply_selection(self, rv, index, is_selected):
        super(GridLocationRVEntry, self).apply_selection(rv, index, is_selected)

        if is_selected:
            self.host_screen.grid_location_selected = rv.data[self.index]['name']
            self.host_screen.set_compatible_apps()
            self.host_screen.set_default_inputs_app_and_size()
            self.host_screen.ids.app_names_rv.refresh_from_data()
            self.host_screen.ids.system_size_rv.refresh_from_data()


class ApplicationRVEntry(RecycleViewRow):
    """The representation for data entries in the RecycleView for 'Application'."""

    def apply_selection(self, rv, index, is_selected):
        super(ApplicationRVEntry, self).apply_selection(rv, index, is_selected)

        if is_selected:
            self.host_screen.application_selected = rv.data[self.index]['name']
            self.host_screen.set_default_inputs_duration()
            self.host_screen.ids.discharge_duration_rv.refresh_from_data()


class SystemSizeRVEntry(RecycleViewRow):
    """The representation for data entries in the RecycleView for 'System Size'."""

    def apply_selection(self, rv, index, is_selected):
        super(SystemSizeRVEntry, self).apply_selection(rv, index, is_selected)

        if is_selected:
            self.host_screen.system_size_selected = rv.data[self.index]['name']


class DischargeDurationRVEntry(RecycleViewRow):
    """The representation for data entries in the RecycleView for 'Discharge Duration'."""

    def apply_selection(self, rv, index, is_selected):
        super(DischargeDurationRVEntry, self).apply_selection(rv, index, is_selected)

        if is_selected:
            self.host_screen.discharge_duration_selected = rv.data[self.index]['name']
            self.host_screen.has_selection = self.host_screen.is_user_selection_complete()


from __future__ import absolute_import

import logging
from datetime import datetime
import calendar
import pyutilib

from kivy.clock import mainthread

from es_gui.tools.valuation.valuation_optimizer import ValuationOptimizer, BadParameterException, IncompatibleDataException


class ValuationOptimizerHandler:
    """A handler for creating and solving ValuationOptimizer instances as requested."""
    dms = None
    solved_ops = []

    def __init__(self, solver_name):
        self._solver_name = solver_name

    @property
    def solver_name(self):
        """The name of the solver for Pyomo to call."""
        return self._solver_name

    @solver_name.setter
    def solver_name(self, value):
        self._solver_name = value

    def process_requests(self, requests, *args):
        """Generates and solves ValuationOptimizer models based on the given requests."""
        dms = self.dms

        iso = requests['iso']
        market_type = requests['market type']
        node_id = str(requests['node id'])
        node_name = self.dms.get_node_name(node_id, iso)
        param_set = requests.get('param set', [None])

        solved_requests = []

        handler_status = set()

        for month, year in requests['months']:
            param_set_iterator = iter(param_set)
            continue_param_loop = True

            while continue_param_loop:
                try:
                    params = next(param_set_iterator)
                except StopIteration:
                    break

                op = ValuationOptimizer(market_type=market_type)

                if iso == 'PJM':
                    #lmp_da, RUP, RDW, MR, RA, RD, RegCCP, RegPCP = dms.get_pjm_data(year, month, node_name)
                    lmp_da, MR, RA, RD, RegCCP, RegPCP = dms.get_pjm_data(year, month, node_id)

                    op.price_electricity = lmp_da
                    op.mileage_mult = MR
                    # op.mileage_slow = RA
                    # op.mileage_fast = RD
                    op.price_regulation = RegCCP
                    op.price_reg_service = RegPCP
                    #op.fraction_reg_up = RUP
                    #op.fraction_reg_down = RDW
                elif iso == 'ERCOT':
                    lmp_da, rd, ru = dms.get_ercot_data(year, month, node_name)

                    op.price_electricity = lmp_da
                    op.price_reg_up = ru
                    op.price_reg_down = rd
                elif iso == 'MISO':
                    lmp_da, regMCP = dms.get_miso_data(year, month, node_name)

                    op.price_electricity = lmp_da
                    # op.price_reg_service = regMCP
                    op.price_regulation = regMCP
                elif iso == 'ISONE':
                    daLMP, RegCCP, RegPCP, miMULT = dms.get_isone_data(year, month, node_id)

                    op.price_electricity = daLMP
                    op.price_regulation = RegCCP
                    op.price_reg_service = RegPCP
                    op.mileage_mult = miMULT
                ########################################################################################################
                elif iso == 'NYISO':
                    lbmp_da, rcap_da = dms.get_nyiso_data(year, month, node_id)

                    op.price_electricity = lbmp_da
                    op.price_regulation = rcap_da
                elif iso == 'SPP':
                    lmp_da, mcpru_da, mcprd_da = dms.get_spp_data(year, month, node_name)

                    op.price_electricity = lmp_da
                    op.price_reg_up = mcpru_da
                    op.price_reg_down = mcprd_da
                elif iso == 'CAISO':
                    lmp_da, aspru_da, asprd_da, asprmu_da, asprmd_da, rmu_mm, rmd_mm, rmu_pacc, rmd_pacc = dms.get_caiso_data(year, month, node_name)

                    op.price_electricity = lmp_da
                    op.price_reg_up = aspru_da
                    op.price_reg_down = asprd_da
                    op.price_reg_serv_up = asprmu_da
                    op.price_reg_serv_down = asprmd_da
                    op.mileage_mult_ru = rmu_mm
                    op.mileage_mult_rd = rmd_mm
                    op.perf_score_ru = rmu_pacc # TODO: give the option to the user to override this
                    op.perf_score_rd = rmd_pacc
                else:
                    logging.error('ValOp Handler: Invalid ISO provided.')
                    raise ValueError('Invalid ISO provided to ValuationOptimizer handler.')

                if params:
                    op.set_model_parameters(**params)
                else:
                    continue_param_loop = False

                try:
                    solved_op = self._solve_model(op)
                except pyutilib.common._exceptions.ApplicationError as e:
                    logging.error('Op Handler: {error}'.format(error=e))

                    if 'No executable found' in e.args[0]:
                        # Could not locate solver executable
                        handler_status.add('* The executable for the selected solver could not be found; please check your installation.')
                    else:
                        handler_status.add('* ({0} {1}) {2}. The problem may be infeasible.'.format(month, year, e.args[0]))
                except IncompatibleDataException as e:
                    # Data exception raised by ValuationOptimizer
                    logging.error(e)
                    handler_status.add('* ({0} {1}) The time series data has mismatched sizes.'.format(month, year))
                except AssertionError as e:
                    # An optimal solution could not be found as reported by the solver
                    logging.error('Op Handler: {error}'.format(error=e))
                    handler_status.add('* ({0} {1}) An optimal solution could not be found; the problem may be infeasible.'.format(month, year))
                else:
                    solved_op = self._save_to_solved_ops(solved_op, iso, market_type, node_name,
                                                        year, month, params)
                    solved_requests.append(solved_op)

        logging.info('Op Handler: Finished processing requested jobs.')
        return solved_requests, handler_status

    def _solve_model(self, op):
        op.solver = self.solver_name
        op.run()

        return op

    @staticmethod
    def _save_to_solved_ops(op, iso, market_type, node_name, year, month, param_set):
        # time_finished = datetime.now().strftime('%A, %B %d, %Y %H:%M:%S')
        time_finished = datetime.now().strftime('%b %d, %Y %H:%M:%S')
        name = ' | '.join([time_finished, node_name, year, calendar.month_abbr[int(month)], repr(param_set)])

        results_dict = {}

        results_dict['name'] = name
        results_dict['optimizer'] = op
        results_dict['iso'] = iso
        results_dict['market type'] = market_type
        results_dict['year'] = year
        results_dict['month'] = calendar.month_name[int(month)]
        results_dict['node'] = node_name
        if param_set:
            results_dict['params'] = param_set
        results_dict['time'] = time_finished
        results_dict['label'] = ' '.join([node_name, year, calendar.month_abbr[int(month)], repr(param_set)])

        ValuationOptimizerHandler.solved_ops.append(results_dict)

        return (name, op)
    
    def get_solved_ops(self):
        """Returns the list of solved Optimizer objects in reverse chronological order."""
        return_list = reversed(self.solved_ops)

        return return_list

if __name__ == '__main__':
    with open('valuation_optimizer.log', 'w'):
        pass

    logging.basicConfig(filename='valuation_optimizer.log', format='[%(levelname)s] %(asctime)s: %(message)s',
                        level=logging.INFO)
    
    op = ValuationOptimizer()

    fpath = os.path.join('data', 'PJM')
    year = 2015
    month = 3
    nodeid = '51217'

    daLMP, mr, rega, regd, RegCCP, RegPCP = read_pjm_data(fpath,year,month,nodeid)

    handler_requests = {}
    handler_requests['iso'] = 'PJM'
    handler_requests['market type'] = 'pjm_pfp'
    handler_requests['months'] = [(month, year),]
    handler_requests['node id'] = nodeid


    results, gross_revenue = op.run()



"""Data Manager for Valuation Tool."""

from __future__ import absolute_import
import os

from quest_eras.quest_library.utilities.pomo.dms import DataManagementSystem
import quest_eras.quest_library.utilities.pomo.valuation.isone.data_processing_tools as dpt


class ValuationDMSISONE(DataManagementSystem):
    """
    A class for managing data for the energy storage valuation optimization functions.

    Class methods for each type of file to be loaded are included, extending from the get_data() method of the superclass.
    Each of these methods uses get_data() to retrieve the relevant data and loads the file and adds it to the DMS if the data is not loaded.
    An optional class method for calling each of the individual data methods can be included to form the necessary arguments and return the desired variables.

    :param home_path: A string indicating the relative path to where data is saved.
    """

    def __init__(self, home_path, **kwargs):
        DataManagementSystem.__init__(self, **kwargs)

        self.home_path = home_path
        self.delimiter = " @ "  # delimiter used to split information in id_key

    def get_node_name(self, node_id, ISO):
        """
        Retrieve the node name corresponding to the given node_id using the lookup table loaded during initialization.

        :param node_id: A string or int of a node ID.
        :return: The corresponding node name as a string.
        """
        # TODO: map node_id to node name

        return str(node_id)

    def get_isone_data(self, year, month, nodeid):
        """Retrieve data for ISO-NE."""
        path = os.path.join(self.home_path, "ISONE")

        year = str(year)
        month = str(month)

        lmp_key = self.delimiter.join([path, year, month, nodeid, "LMP"])
        rccp_key = self.delimiter.join([path, year, month, "RegCCP"])
        rpcp_key = self.delimiter.join([path, year, month, "RegPCP"])
        mimult_key = self.delimiter.join([path, year, month, "MiMult"])

        try:
            # attempt to access data if it is already loaded
            lmp_da = self.get_data(lmp_key)
            rccp = self.get_data(rccp_key)
            rpcp = self.get_data(rpcp_key)
            mi_mult = self.get_data(mimult_key)
        except KeyError:
            # load the data and add it to the DMS
            lmp_da, rccp, rpcp, mi_mult = dpt.read_isone_data(path, year, month, nodeid)

            self.add_data(lmp_da, lmp_key)
            self.add_data(rccp, rccp_key)
            self.add_data(rpcp, rpcp_key)
            self.add_data(mi_mult, mimult_key)

        return lmp_da, rccp, rpcp, mi_mult


# if __name__ == "__main__":
#     dms = ValuationDMS(save_name="valuation_dms.p", home_path="data")

#     # # ERCOT - data doesn't exist
#     # year = 2010
#     # month = 1
#     # settlement_point = 'HB_HOUSTON'

#     # spp_da, rd, ru = dms.get_ercot_data(year, month, settlement_point)

#     # # ERCOT
#     # year = 2010
#     # month = 12
#     # settlement_point = 'HB_HOUSTON'

#     # spp_da, rd, ru = dms.get_ercot_data(year, month, settlement_point)

#     # PJM
#     # year = 2016
#     # month = 5
#     # nodeid = 1

#     # lmp_da, MR, RA, RD, RegCCP, RegPCP = dms.get_pjm_data(year, month, nodeid)

#     # MISO
#     year = 2015
#     month = 3
#     nodeid = "AEC"

#     lmp_da, RegMCP = dms.get_miso_data(year, month, nodeid)
